
module A = Ast
module Ctx = TypeCheckingContext  
module HS = HereditarySubstitution


(* exception SolutionNotFound
exception FoundSolution  of (Ctx.ctxtype * A.tm)  (* solution*)
                          * ( unit -> unit ) continue *)

let debug = Flags.lp_debug

type focus_mode = FocusedOnRecConst  (* the focus is on rec consts, arguments must be immediate variables *)
                  | FocusedOnVarOnly (* the focus is on single variable, must check if the focused can solve the goal directly*)
                  | FocusedOnBound (* the argument can be anything *)
                
(* let callcc (k : ('a -> 'b) -> 'a ) : 'a = 
  Base.With_return.with_return ( fun r -> k (fun v -> r.return v)) *)


let ds_pair ((tm,tp) : A.tm * A.tp ) : string = 
  A.ds_tm tm ^ " : " ^ A.ds_tp tp
  

let update_tm_tp_pair (ctx : Ctx.ctxtype) ((tm, tp) : A.tm * A.tp) : A.tm * A.tp = 
  (Ctx.instantiate_metavar_tm ctx tm, Ctx.instantiate_metavar_tp ctx tp)

(* returns the term that solves the goal*)
let rec focus_l (ctx : Ctx.ctxtype) (focus_mode : focus_mode) 
   (focus : A.tm * A.tp) (goal : A.tm * A.tp) (g : A.dec list) : (Ctx.ctxtype * A.tm) = 
   (* TODO: implement early failure if the focuse's target type's head is distinct from the goal to speed up back chaining *)
  let _ = if debug then print_endline ("focus_l on " ^ ds_pair focus  ^ " >> " ^ ds_pair goal ^ Ctx.ds_conditional_show_ctx ctx) in
  let (focus, goal) = (update_tm_tp_pair ctx focus, update_tm_tp_pair ctx goal) in
  (* check focus mode on vars agrees with what is provided *)
  let _ = if focus_mode = FocusedOnVarOnly 
              then (match (goal) with
                    | (A.Atm(_, []), _) -> ()
                    | _ -> failwith "focus_l var only with malformed goal term")
              else ()
          in
  let (ftm, ftp) = focus in
    (
      match (ftp, focus_mode) with
      (* base case: focused type is atomic OR focus mode is var only*)
        | ((A.Atp _, _) | (_, FocusedOnVarOnly))   -> 
          let (gtm, gtp) = goal in
          let (ctx, tp) = Unification.request_unify_tp ctx ftp gtp (g, []) in
          let (ctx, tm) = Unification.request_unify_tm ctx ftm gtm tp (g, []) in
          (ctx, tm)
        | (A.Pi (x, t2, t1), _) -> 
          (
              match ftm with
              | A.Lam _ -> failwith "focus cannot be on a lambda"
              | A.Atm(ftmh, ftms) ->
                ( 
                  let should_not_solve_premise = HS.occurs_in_tp x t1 in
                  let (ctx,tm, t1) = Ctx.instantiate_pi_with_new_metavar ctx Ctx.collect_all_bound_vars ftp [] (g, []) in
                  let (ctx, sol) = focus_l ctx focus_mode (A.Atm(ftmh, ftms@[tm]), t1) goal g in
                  if should_not_solve_premise (* TODO : figure out the case for FocusOnRecVars*)
                    then (
                      (ctx, sol)
                    )
                    else (
                      (* update tm*)
                      let tm = Ctx.instantiate_metavar_tm ctx tm in
                      let (ctx, _) = 
                        (match focus_mode with 
                          | FocusedOnBound -> focus_r ctx (tm, t2) g 
                          | FocusedOnRecConst -> choose_focus ctx FocusedOnVarOnly (tm, t2) g
                          | _ -> failwith "Not Possible"
                        ) in
                      let sol = Ctx.instantiate_metavar_tm ctx sol in
                      (ctx, sol)
                    )
                )
          )
        | _ -> failwith "goal cannot be Type/CoType"
    )


and choose_focus (ctx : Ctx.ctxtype) (focus_mode : focus_mode) (goal : A.tm * A.tp)(g : A.dec list) : (Ctx.ctxtype * A.tm )  = 
(* contractive candidates *)
    (* collects assumptions that can be appealed to in the context*) 
    let (ctxvars) = Ctx.collect_all_bound_vars ctx in
    (* collects constants from signatures, the definitions are ignored *)
    let sigvars = List.filter_map (fun v -> match v with
    | A.Decl(c, a, _) -> if not (A.is_kind a) then Some (c, a) else None
    | _ -> None) g in
(* noncontractive focus candidates *)
    let rec_vars = Ctx.collect_all_rec_constants ctx in
    let rec_vars = List.map (fun id -> let (tp, _) = Ctx.lookup_id_in_ctx ctx id in (id, tp)) rec_vars in
    let all_focusl_candidates = 
      (match focus_mode with
        | FocusedOnBound | FocusedOnVarOnly -> ctxvars @ sigvars 
        | FocusedOnRecConst -> rec_vars 
      ) in
    (* simulate call cc with exceptions *)
    let rec choose_focus_inner_loop curIdx = 
      ( if curIdx = List.length all_focusl_candidates 
        then raise (Failure "cannot resolve goal")
        else (try (
          let (fid, ftp) = (List.nth all_focusl_candidates curIdx) in
            focus_l ctx focus_mode (A.Atm(fid, []), ftp) goal g)
              with Failure s -> 
                (
                  if debug then print_endline ("LP debug: subprocedure failed, backtracking, failure is: " ^ s) else ();
                  choose_focus_inner_loop (curIdx + 1)
                )
              )
      )
    in
     choose_focus_inner_loop 0 
     

(* Free variables will be instantiated as metavars *)
(* focus_r always assumes the recursive goals, that is, goals may be unified with some recursion constant*)
and focus_r (ctx : Ctx.ctxtype) ( goal : A.tm * A.tp) (g : A.dec list) : Ctx.ctxtype * A.tm = 
  let _ = if debug then print_endline ("focus_r on  " ^ ds_pair goal ^ Ctx.ds_conditional_show_ctx ctx) in
  let (goal) = (update_tm_tp_pair ctx goal) in
  let (gtm, gtp) = goal in
  (* match gtm with
  | A.Lam _ -> failwith "lp53"
  | A.Atm(h, s) -> *)
    (match gtp with
      | A.Pi (x, t2, t1) -> 
        let fresh_name = A.Prvd(Uid.nextz()) in
        let t1 = HS.rename_in_tp (fresh_name) x t1 in
        (Ctx.with_new_name ctx fresh_name t2 (fun ctx -> 
                let (ctx, b) = Ctx.new_metavar_tm_with_abstraction ctx Ctx.collect_all_bound_vars gtp [] (g, []) in
                let (ctx, _) = Unification.request_unify_tm ctx gtm (A.Lam(x, b)) gtp (g, []) in
                let (ctx, _) =  focus_r ctx (b, t1) g in
                (ctx, (A.Lam(x, b)))
          ))
      | A.Atp (_, _) -> (
        (
            match gtm with
            | A.Atm(h_tm, tm_s) -> 
              (match Ctx.find_id_in_ctx ctx h_tm with
                | Some(_, Ctx.IdUniVar(Ctx.UTmVarRec _)) -> (

                  try (
                    (* first try recursive focus *)
                    choose_focus ctx FocusedOnRecConst goal g
                    )
                  with Failure failure -> 
                    (
                      match h_tm with
                      | A.Gen n1 -> 
                        (
                          if debug then print_endline ("LP debug: subprocedure failed, backtracking, failure is: " ^ failure) else ();
                          (* else create a recursion point and *)
                          let s' = Unification.check_spine_is_pattern ctx tm_s (g, []) in 
                          let (ctx, newUnivar) = Ctx.turn_rec_univar_into_contra_univar ctx n1 s' (g,[]) in
                          let gtm = A.Atm(newUnivar, tm_s) in
                          choose_focus ctx FocusedOnBound (gtm, gtp) g
                        )
                      | _ -> failwith "Univar must be Gen"
                    )
                )
                | _ -> failwith ("CHECK: Why the target of a focus NOT a recursive unification variable?, but is " ^ A.ds_tm gtm)
              )
            | _ -> failwith "atomic goals have lambda terms as its solution, not possible"
        )
      )
      | _ -> failwith "cannot focus R on Type/CoType, is a kind not a type"
    )
    

let rec focus_r_top_level (ctx : Ctx.ctxtype) (goal : A.tp) 
(free_var_assoc : (A.id * A.tm) list  (* associates names of free var to unification vars*))
 (free_vars_as_metavars : A.id list) (g : A.dec list) : (Ctx.ctxtype * ((A.id * A.tm) list * A.tm)) = 
  let _ = if debug then print_endline ("focus_r_top_level on  " ^ A.ds_tp goal ^ Ctx.ds_conditional_show_ctx ctx) in
    match goal with
    | A.Pi (x, t2, t1) -> 
      if List.mem (x) free_vars_as_metavars 
        then (let (ctx, tm, t1) = Ctx.instantiate_pi_with_new_metavar ctx Ctx.collect_all_bound_vars goal [] (g, []) in
              focus_r_top_level ctx t1 (free_var_assoc@[(x, tm)]) free_vars_as_metavars g 
              )
        else (Ctx.with_new_name ctx x t2 (fun ctx -> 
                let (ctx,(assoc, b)) =  focus_r_top_level ctx t1 free_var_assoc free_vars_as_metavars g in
                (ctx,(assoc, A.Lam(x, b)))
          ))
    | A.Atp _ -> 
      let (ctx, tm) = Ctx.new_metavar_tm_with_abstraction ctx (Ctx.collect_all_bound_vars) goal  [] (g, []) in
      let (ctx, _) = focus_r ctx (tm, goal) g in
      let tm_final = Ctx.instantiate_metavar_tm ctx tm in
      (ctx, (free_var_assoc, tm_final))
    | _ -> failwith "top level goals cannot have type/cotype at the end"


let rec loop ( g : A.dec list): unit = 
  try(
  let _ = print_endline "" in
  let _  = Stdlib.output_string Stdlib.stdout "?- "  in
  let _ = Stdlib.flush Stdlib.stdout  in
  let input = Stdlib.input_line Stdlib.stdin in
  let processed = Lexer.lex_string input in
  let rawparsed = Parse.rawparse processed in
  let elaborated = Elborate.elaborateDec rawparsed in
  match elaborated with
  | [A.Decl(name, tp, None)] -> 
      (
        (
          let _ = print_endline ("Read as " ^ A.ds_tp tp) in
          let (free_vars) = Ctx.get_capital_free_vars_in_tp tp (g, []) in
          let _ = print_endline ("Existential Pi's are " ^ (String.concat ", " (List.map A.ds_id free_vars))) in
          let (_, reconstructed, _) = TypeReconstruct.check_kind_or_type tp (g, [])  in
          let _ = print_endline ( "Reconstructed as " ^ A.ds_tp reconstructed) in
          let ctx = [] in
          let ctx = Ctx.populate_ctx_with_rec_constants_from_signature ctx g in
          (* let ctx_initial = ctx in *)
          let (ctx, (assoc, tm)) = focus_r_top_level ctx reconstructed [] ( free_vars) g in
          (* let additional_rec_const_ids =
            ListUtil.minus (Ctx.collect_all_rec_constants ctx)
                (Ctx.collect_all_rec_constants ctx_initial) in *)
          (* let additional_rec_const_ids = Ctx.collect_all_rec_constants ctx in *)
          let unif_sol : (A.id * A.tm) list = (List.map (fun fv ->
              let (_, tm) = List.find (fun (fvname, _) -> (fv) = fvname) assoc in
              let tm = Ctx.instantiate_metavar_tm ctx tm in
              (fv, tm)
            ) free_vars) in
          let unif_sol_text = String.concat "\n" (List.map (fun (fv, tm) ->
              A.ds_id fv ^ " = " ^ A.ds_tm tm
            ) unif_sol) in
          let query_sol = A.ds_id name ^ " = " ^ A.ds_tm tm in
          let additional_rec_const_ids = CircularConversion.efficiently_represent_tms ctx (tm:: (List.map (fun (_, m) -> m) unif_sol)) in
          let additional_recs = String.concat "\n" (List.map (fun (recid, tp, tm) -> 
                  A.ds_id recid ^ " : " ^ A.ds_tp tp ^ " = " ^ A.ds_tm tm
            ) additional_rec_const_ids) in
          let _ = Stdlib.output_string Stdlib.stdout (additional_recs ^ "\n" ^ unif_sol_text ^ "\n" ^ query_sol) in
          loop g
          )
      )
  | _ -> failwith "lp22: expect single declaration in the form of a : A."
  ) with 
   Failure s -> (print_endline s; print_endline "Exception occurred during program execution. Try again."; loop g)
  
  
  