
module P = Parseast
module A = Ast


let priority_counter = ref 10000
let nextPriority() : int = 
  let _ = priority_counter := !priority_counter - 1 in
   !priority_counter 

let elaborateId (p : P.parseast) : A.id = 
  match p with
  | P.Id(Str s) -> A.Prvd s
  | P.Id(Placeholder) -> A.Placeholder
  | _ -> failwith ("Cannot elaborate id " ^ P.debugshow p)


let rec elaborateTm ( p : P.parseast) : A.tm = 
  match p with
  | P.App _ -> let (h, s) = (elaborateHeadSpine p) in
              A.Atm(h, s)
  | P.Id _ -> A.Atm(elaborateId p, [])
  | P.LamT(_) -> failwith "does not support typed lambda abstraction, please remove types"
  (* | P.LamT(x, t2, m1) -> A.Lam(elaborateId x, elaborateTp t2, elaborateTm m1) *)
  | P.LamNT(x, m1) -> A.Lam(elaborateId x, elaborateTm m1)
  | P.Paren(p) -> elaborateTm p
  | _ -> failwith ("Cannot elaborate tm at " ^ P.debugshow p)
and elaborateHeadSpine ( p : P.parseast) : A.id * A.tm list = 
    match p with
    | P.Id _ -> (elaborateId p, [])
    | P.App(t1, m2) -> let (h, s) = elaborateHeadSpine t1 in (h, s @[elaborateTm m2])
    | _ -> failwith ("cannot elaborate headspine " ^ P.debugshow p)

and elaborateTp (p : P.parseast) : A.tp = 
  match p with
  | P.PiT(x, t2, t1) -> A.Pi(elaborateId x, elaborateTp t2, elaborateTp t1)
  | P.PiNT(x, t1) -> A.Pi(elaborateId x, A.Atp(A.Placeholder, []), elaborateTp t1)
  | P.PiArr(t2, t1) -> A.Pi(Placeholder, elaborateTp t2, elaborateTp t1)
  | P.Id(P.Str "type")-> A.Type (nextPriority())
  | P.Id(P.Str "cotype")-> A.CoType (nextPriority())
  | P.Id _ -> A.Atp(elaborateId p, [])
  | P.App _ -> let (h, s) = elaborateHeadSpine p in A.Atp(h, s)
  | P.Paren(p) -> elaborateTp p
  | _ -> failwith ("cannot elaborate type " ^ P.debugshow p)





let rec elaborateDec (p : P.parseast) : A.dec list = 
(* let _ = print_string ("elaborating" ^ P.show_parseast p ^ "\n") in *)
  match p with
  | Dec2(d1, d2) -> 
    let e1 = elaborateDec d1 in let e2 =  elaborateDec d2 in e1 @ e2
  | Dec(c, a) -> [A.Decl(elaborateId c, elaborateTp a, None)]
  | DecDefC(c, a, m) -> [A.Def(elaborateId c, elaborateTp a, elaborateTm m, A.Unchecked)] (*TODO: separate them is the step after next*)
  | DecDefI(c, a, m) -> [A.Def(elaborateId c, elaborateTp a, elaborateTm m, A.Unchecked)]
  | _ -> failwith ("Cannot elaborate delcaration at ^ " ^ P.debugshow p)